/* StarfieldEffect.ino
  by Adam Wolf

  Uses a SmartMatrix SD Shield, a 32 by 32 RGB LED display, and a
  Teensy to show a starfield effect.
*/

#include <SmartMatrix3.h>
#define COLOR_DEPTH 24
const uint8_t kMatrixWidth = 32;
const uint8_t kMatrixHeight = 32;
const uint8_t kRefreshDepth = 36;
const uint8_t kDmaBufferRows = 4;
const uint8_t kPanelType = SMARTMATRIX_HUB75_32ROW_MOD16SCAN;
const uint8_t kMatrixOptions = (SMARTMATRIX_OPTIONS_NONE);
const uint8_t kBackgroundLayerOptions = (SM_BACKGROUND_OPTIONS_NONE);
const uint8_t kScrollingLayerOptions = (SM_SCROLLING_OPTIONS_NONE);
const uint8_t kIndexedLayerOptions = (SM_INDEXED_OPTIONS_NONE);
SMARTMATRIX_ALLOCATE_BUFFERS(matrix, kMatrixWidth, kMatrixHeight,
                             kRefreshDepth, kDmaBufferRows, kPanelType, kMatrixOptions);
SMARTMATRIX_ALLOCATE_BACKGROUND_LAYER(backgroundLayer, kMatrixWidth,
                                      kMatrixHeight, COLOR_DEPTH, kBackgroundLayerOptions);
SMARTMATRIX_ALLOCATE_SCROLLING_LAYER(scrollingLayer, kMatrixWidth,
                                     kMatrixHeight, COLOR_DEPTH, kScrollingLayerOptions);
SMARTMATRIX_ALLOCATE_INDEXED_LAYER(indexedLayer, kMatrixWidth,
                                   kMatrixHeight, COLOR_DEPTH, kIndexedLayerOptions);

struct Star {
  int16_t x;
  int16_t y;
  float z;
};

#define NUM_STARS 32
struct Star stars[NUM_STARS];

const int X_RANGE = 400;
const int Y_RANGE = 400;
const int HALF_X_RANGE = X_RANGE / 2;
const int HALF_Y_RANGE = Y_RANGE / 2;
const int MAX_Z = 16;

const uint8_t SCREEN_WIDTH = 32;
const uint8_t SCREEN_HEIGHT = 32;
const uint8_t HALF_SCREEN_WIDTH = SCREEN_WIDTH / 2;
const uint8_t HALF_SCREEN_HEIGHT = SCREEN_HEIGHT / 2;

const float SPEED = 0.1;

void randomizeStarPosition(struct Star* s)
{
  s->x = random(-HALF_X_RANGE, HALF_X_RANGE);
  s->y = random(-HALF_Y_RANGE, HALF_Y_RANGE);
  s->z = MAX_Z; // Put the star in the back.
}

void drawStar(struct Star* s)
{
  uint8_t display_x = (int) (s->x / s->z);
  uint8_t display_y = (int) (s->y / s->z);

  display_x = display_x + HALF_SCREEN_WIDTH;
  display_y = display_y + HALF_SCREEN_HEIGHT;

  //Let's check if our calculated point should be displayed.
  if (display_x >= 0 &&
      display_y >= 0 &&
      display_x < SCREEN_WIDTH &&
      display_y < SCREEN_HEIGHT)
  {
    uint8_t brightness = map(s->z, 0, MAX_Z, 255, 0);
    drawPoint(display_x, display_y, brightness); // It fits, so draw it!
  } else
  {
    randomizeStarPosition(s);  // It does not fit, so make a new star.
  }
}

void advanceStar(struct Star* s)
{
  s->z -= SPEED;

  if (s->z <= 0)
  {
    randomizeStarPosition(s);  //If the star has passed us, make a new one.
  }
}

void setup()
{
  randomSeed(analogRead(A1));

  for (int i = 0; i < NUM_STARS; i++)
  {
    Star* current_star = &stars[i];
    randomizeStarPosition(current_star);
    current_star->z = random(0, MAX_Z);
  }

  setupDisplay();
}


void loop() {
  clearDisplay();

  for (int i = 0; i < NUM_STARS; i++)
  {
    drawStar(&stars[i]);
  }

  paintDisplay();

  for (int i = 0; i < NUM_STARS; i++)
  {
    advanceStar(&stars[i]);
  }
}

uint16_t distanceToStar(struct Star* s)
{
  return sqrt(sq(s->x) + sq(s->y) + sq(s->z)); // distance from (0, 0, 0)
}

void clearDisplay()
{
  backgroundLayer.fillScreen({0x00, 0x00, 0x00});
}

void drawPoint(uint8_t x, uint8_t y, uint8_t brightness)
{
  // Note this doesn't actually show it on the display.
  backgroundLayer.drawPixel(x, y, {brightness, brightness, brightness});
}

void paintDisplay()
{
  // Send the image we've built up to the screen.
  backgroundLayer.swapBuffers(false);
}

void setupDisplay()
{
  matrix.addLayer(&backgroundLayer);
  matrix.addLayer(&scrollingLayer);
  matrix.addLayer(&indexedLayer);
  matrix.begin();

  backgroundLayer.enableColorCorrection(true);
}

